# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import sys

from oslo_log import log as logging

from networking_mlnx.eswitchd.common import constants
from networking_mlnx.eswitchd.db import eswitch_db
from networking_mlnx.eswitchd.resource_mngr import ResourceManager
from networking_mlnx.eswitchd.utils import command_utils
from networking_mlnx.eswitchd.utils import pci_utils


LOG = logging.getLogger(__name__)

INVALID_PKEY = 'none'
DEFAULT_PKEY_IDX = '0'
PARTIAL_PKEY_IDX = '1'
DEFAULT_MASK = 0x7fff
DEFAULT_PKEY = '0xffff'


class eSwitchHandler(object):

    def __init__(self, fabrics=None):
        self.eswitches = {}
        self.pci_utils = pci_utils.pciUtils()
        self.rm = ResourceManager()
        self.devices = set()
        if fabrics:
            self.add_fabrics(fabrics)

    def add_fabrics(self, fabrics):
        for fabric, pf in fabrics:
            verify_vendor_pf = (
                self.pci_utils.verify_vendor_pf(pf, constants.VENDOR))
            if (not verify_vendor_pf or
                not self.pci_utils.is_sriov_pf(pf)):
                LOG.error("PF %s must have Mellanox Vendor ID "
                          ",SR-IOV and driver module "
                          "enabled. Terminating!", pf)
                sys.exit(1)

            if self.eswitches.get(fabric) is None:
                self.eswitches[fabric] = []
            vfs = self.pci_utils.get_vfs_info(pf)
            self.eswitches[fabric].append(
                eswitch_db.eSwitchDB(pf=pf, vfs=vfs))
            self._add_fabric(fabric, pf)

        self.sync_devices()

    def sync_devices(self):
        devices, vm_ids = self.rm.scan_attached_devices()
        added_devs = {}
        removed_devs = {}
        added_devs = set(devices) - self.devices
        removed_devs = self.devices - set(devices)
        self._treat_added_devices(added_devs, vm_ids)
        self._treat_removed_devices(removed_devs)
        self.devices = set(devices)

    def _add_fabric(self, fabric, pf):
        self.rm.add_fabric(fabric, pf)
        self._config_port_up(pf)
        pf_fabric_details = self.rm.get_fabric_details(fabric, pf)
        eswitches = self._get_eswitches_for_fabric(fabric)
        eswitch = None
        for esw in eswitches:
            if esw.pf == pf:
                eswitch = esw
                break
        for vf in pf_fabric_details['vfs']:
            eswitch.create_port(vf, constants.VIF_TYPE_HOSTDEV)

    def _treat_added_devices(self, devices, vm_ids):
        for device in devices:
            dev, mac, fabric = device
            if fabric:
                for eswitch in self.eswitches[fabric]:
                    if dev in eswitch.vfs:
                        eswitch.attach_vnic(
                            port_name=dev, device_id=vm_ids[dev], vnic_mac=mac)
                        if eswitch.vnic_exists(mac):
                            eswitch.plug_nic(port_name=dev)
                        break
            else:
                LOG.info("No Fabric defined for device %s", dev)

    def _treat_removed_devices(self, devices):
        for dev, mac in devices:
            fabric = self.rm.get_fabric_for_dev(dev)
            if fabric:
                for eswitch in self.eswitches[fabric]:
                    if dev in eswitch.vfs:
                        eswitch.detach_vnic(vnic_mac=mac)
            else:
                LOG.info("No Fabric defined for device %s", dev)

    def get_vnics(self, fabrics):
        vnics = {}
        for fabric in fabrics:
            eswitches = self._get_eswitches_for_fabric(fabric)
            if eswitches:
                for eswitch in eswitches:
                    vnics_for_eswitch = eswitch.get_attached_vnics()
                    vnics.update(vnics_for_eswitch)
            else:
                LOG.error("No eSwitch found for Fabric %s", fabric)
                continue
        LOG.info("vnics are %s", vnics)
        return vnics

    def plug_nic(self, fabric, device_id, vnic_mac, pci_slot):
        eswitch = self._get_eswitch_for_fabric_and_pci(fabric, pci_slot)
        if eswitch:
            eswitch.port_table[pci_slot]['vnic'] = vnic_mac
            eswitch.port_policy.update(
                {vnic_mac:
                    {'vlan': None,
                     'dev': pci_slot,
                     'device_id': device_id}})

            self._config_vf_mac_address(fabric, pci_slot, vnic_mac)
            eswitch.plug_nic(pci_slot)
        else:
            LOG.error("No eSwitch found for Fabric %s", fabric)

        return pci_slot

    def delete_port(self, fabric, vnic_mac):
        dev = None
        eswitches = self._get_eswitches_for_fabric(fabric)
        for eswitch in eswitches:
            if eswitch and vnic_mac in eswitch.get_attached_vnics():
                dev = eswitch.detach_vnic(vnic_mac)
                if dev:
                    self._config_vf_mac_address(fabric, dev)
                    break

        if dev is None:
            LOG.warning("No eSwitch found for Fabric %s", fabric)
        return dev

    def port_release(self, fabric, vnic_mac):
        ret = None
        dev = None
        eswitches = self._get_eswitches_for_fabric(fabric)
        for eswitch in eswitches:
            if eswitch and vnic_mac in eswitch.get_attached_vnics():
                dev = eswitch.get_dev_for_vnic(vnic_mac)
                if (dev is not None and eswitch.get_port_state(dev) ==
                    constants.VPORT_STATE_UNPLUGGED):
                    ret = self.set_vlan(
                        fabric, vnic_mac, constants.UNTAGGED_VLAN_ID)
                    self.port_down(fabric, vnic_mac)
                eswitch.port_release(vnic_mac)
        return ret

    def port_up(self, fabric, vnic_mac):
        dev = None
        eswitches = self._get_eswitches_for_fabric(fabric)
        for eswitch in eswitches:
            if eswitch and vnic_mac in eswitch.get_attached_vnics():
                dev = eswitch.get_dev_for_vnic(vnic_mac)
                break
        if not dev:
            LOG.info("No device for MAC %s", vnic_mac)

    def port_down(self, fabric, vnic_mac):
        dev = None
        eswitches = self._get_eswitches_for_fabric(fabric)
        for eswitch in eswitches:
            if eswitch and vnic_mac in eswitch.get_attached_vnics():
                dev = eswitch.get_dev_for_vnic(vnic_mac)
                if dev:
                    LOG.info("IB port for MAC %s doen't support "
                             "port down", vnic_mac)
                break
        if dev is None:
            LOG.info("No device for MAC %s", vnic_mac)

    def set_vlan(self, fabric, vnic_mac, vlan):
        eswitches = self._get_eswitches_for_fabric(fabric)
        for eswitch in eswitches:
            if eswitch and vnic_mac in eswitch.get_attached_vnics():
                eswitch.set_vlan(vnic_mac, vlan)
                dev = eswitch.get_dev_for_vnic(vnic_mac)
                state = eswitch.get_port_state(dev)
                if dev:
                    if state in (constants.VPORT_STATE_ATTACHED,
                                 constants.VPORT_STATE_UNPLUGGED):
                        if eswitch.get_port_table()[dev]['alias']:
                            dev = eswitch.get_port_table()[dev]['alias']
                        try:
                            self._config_vlan_ib(fabric, dev, vlan)
                            return True
                        except RuntimeError:
                            LOG.error('Set VLAN operation failed')
        return False

    def get_eswitch_tables(self, fabrics):
        tables = {}
        for fabric in fabrics:
            eswitches = self._get_eswitches_for_fabric(fabric)
            if len(eswitches) == 0:
                LOG.info("Get eswitch tables: No eswitch %s", fabric)
                continue
            for eswitch in eswitches:
                if eswitch:
                    tables[fabric] = {
                        'port_table': eswitch.get_port_table_matrix(),
                        'port_policy': eswitch.get_port_policy_matrix()
                    }
        return tables

    def _get_eswitches_for_fabric(self, fabric):
        if fabric in self.eswitches:
            return self.eswitches[fabric]
        else:
            return

    def _get_eswitch_for_fabric_and_pci(self, fabric, pci_slot):
        eswitches = self._get_eswitches_for_fabric(fabric)
        for eswitch in eswitches:
            if pci_slot in eswitch.vfs:
                return eswitch

    def _config_vf_pkey(self, ppkey_idx, pkey_idx,
                        pf_mlx_dev, vf_pci_id, hca_port):
        path = constants.MLNX4_PKEY_INDEX_PATH % (pf_mlx_dev, vf_pci_id,
                                                  hca_port, pkey_idx)
        cmd = ['ebrctl', 'write-sys', path, ppkey_idx]
        command_utils.execute(*cmd)

    def _get_guid_idx(self, pf_mlx_dev, dev, hca_port):
        path = constants.MLNX4_GUID_INDEX_PATH % (pf_mlx_dev, dev, hca_port)
        with open(path) as fd:
            idx = fd.readline().strip()
        return idx

    def _get_guid_from_mac(self, mac, device_type):
        guid = None
        if device_type == constants.MLNX4_DEVICE_TYPE:
            if mac is None:
                guid = constants.MLNX4_INVALID_GUID
            else:
                mac = mac.replace(':', '')
                prefix = mac[:6]
                suffix = mac[6:]
                guid = prefix + '0000' + suffix
        elif (device_type == constants.MLNX5_DEVICE_TYPE):
            if mac is None:
                guid = constants.MLNX5_INVALID_GUID
            else:
                prefix = mac[:9]
                suffix = mac[9:]
                guid = prefix + '00:00:' + suffix
        return guid

    def _config_vf_mac_address(self, fabric, dev, vnic_mac=None):
        pf_fabric_details = self._get_pf_fabric(fabric, dev)
        vf_device_type = pf_fabric_details['vfs'][dev]['vf_device_type']
        vguid = self._get_guid_from_mac(vnic_mac, vf_device_type)
        if vf_device_type == constants.MLNX4_DEVICE_TYPE:
            self._config_vf_mac_address_mlnx4(vguid, dev, pf_fabric_details)
        elif (vf_device_type == constants.MLNX5_DEVICE_TYPE):
            self._config_vf_mac_address_mlnx5(vguid, dev, pf_fabric_details)
        else:
            LOG.error("Unsupported vf device type: %s ", vf_device_type)

    def _config_vf_mac_address_mlnx4(self, vguid, dev, pf_fabric_details):
        hca_port = pf_fabric_details['hca_port']
        pf_mlx_dev = pf_fabric_details['pf_mlx_dev']
        self._config_vf_pkey(
            INVALID_PKEY, DEFAULT_PKEY_IDX, pf_mlx_dev, dev, hca_port)

        guid_idx = self._get_guid_idx(pf_mlx_dev, dev, hca_port)
        path = constants.MLNX4_ADMIN_GUID_PATH % (
            pf_mlx_dev, hca_port, guid_idx)
        cmd = ['ebrctl', 'write-sys', path, vguid]
        command_utils.execute(*cmd)
        ppkey_idx = self._get_pkey_idx(
            int(DEFAULT_PKEY, 16), pf_mlx_dev, hca_port)
        if ppkey_idx >= 0:
            self._config_vf_pkey(
                ppkey_idx, PARTIAL_PKEY_IDX, pf_mlx_dev, dev, hca_port)
        else:
            LOG.error("Can't find partial management pkey for "
                      "%(pf)s:%(dev)s", {'pf': pf_mlx_dev, 'dev': dev})

    def _config_vf_mac_address_mlnx5(self, vguid, dev, pf_fabric_details):
        vf_num = pf_fabric_details['vfs'][dev]['vf_num']
        pf_mlx_dev = pf_fabric_details['pf_mlx_dev']
        guid_node = constants.MLNX5_GUID_NODE_PATH % {'module': pf_mlx_dev,
                                                      'vf_num': vf_num}
        guid_port = constants.MLNX5_GUID_PORT_PATH % {'module': pf_mlx_dev,
                                                      'vf_num': vf_num}
        guid_poliy = constants.MLNX5_GUID_POLICY_PATH % {'module': pf_mlx_dev,
                                                         'vf_num': vf_num}
        for path in (guid_node, guid_port):
            cmd = ['ebrctl', 'write-sys', path, vguid]
            command_utils.execute(*cmd)

        if vguid == constants.MLNX5_INVALID_GUID:
            cmd = ['ebrctl', 'write-sys', guid_poliy, 'Down\n']
            command_utils.execute(*cmd)
            cmd = ['ebrctl', 'write-sys', constants.UNBIND_PATH, dev]
            command_utils.execute(*cmd)
            cmd = ['ebrctl', 'write-sys', constants.BIND_PATH, dev]
            command_utils.execute(*cmd)
        else:
            cmd = ['ebrctl', 'write-sys', guid_poliy, 'Up\n']
            command_utils.execute(*cmd)

    def _config_vlan_ib(self, fabric, dev, vlan):
        pf_fabric_details = self._get_pf_fabric(fabric, dev)
        hca_port = pf_fabric_details['hca_port']
        pf_mlx_dev = pf_fabric_details['pf_mlx_dev']
        vf_device_type = pf_fabric_details['vfs'][dev]['vf_device_type']
        if vf_device_type == constants.MLNX4_DEVICE_TYPE:
            self._config_vlan_ib_mlnx4(vlan, pf_mlx_dev, dev, hca_port)
        elif vf_device_type == constants.MLNX5_DEVICE_TYPE:
            pass
        else:
            LOG.error("Unsupported vf device type: %s ", vf_device_type)

    def _config_vlan_ib_mlnx4(self, vlan, pf_mlx_dev, dev, hca_port):
        if vlan == 0:
            ppkey_idx = self._get_pkey_idx(
                int(DEFAULT_PKEY, 16), pf_mlx_dev, hca_port)
            if ppkey_idx >= 0:
                self._config_vf_pkey(
                    ppkey_idx, DEFAULT_PKEY_IDX, pf_mlx_dev, dev, hca_port)
        else:
            ppkey_idx = self._get_pkey_idx(str(vlan), pf_mlx_dev, hca_port)
            if ppkey_idx:
                self._config_vf_pkey(
                    ppkey_idx, DEFAULT_PKEY_IDX, pf_mlx_dev, dev, hca_port)

    def _get_pkey_idx(self, vlan, pf_mlx_dev, hca_port):
        PKEYS_PATH = "/sys/class/infiniband/%s/ports/%s/pkeys/*"
        paths = PKEYS_PATH % (pf_mlx_dev, hca_port)
        for path in glob.glob(paths):
            fd = open(path)
            pkey = fd.readline()
            fd.close()
            # the MSB in pkey is the membership bit ( 0 - partial, 1 - full)
            # the other 15 bit are the number of the pkey
            # so we want to remove the 16th bit when compare pkey file
            # to the vlan (pkey) we are looking for
            is_match = int(pkey, 16) & DEFAULT_MASK == int(vlan) & DEFAULT_MASK
            if is_match:
                return path.split('/')[-1]
        return None

    def _config_port_up(self, dev):
        cmd = ['ip', 'link', 'set', dev, 'up']
        command_utils.execute(*cmd)

    def _get_pf_fabric(self, fabric, dev):
        fabric_details = self.rm.get_fabric_details(fabric)
        for pf_fabric in fabric_details.values():
            if dev in pf_fabric['vfs']:
                return pf_fabric



# Copyright 2016 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""sdn_journal change data to text

Revision ID: 5d5e04ea01d5
Create Date: 2016-08-16 06:01:54.795542

"""

from alembic import op
import sqlalchemy as sa

from networking_mlnx.plugins.ml2.drivers.sdn import constants as sdn_const


# revision identifiers, used by Alembic.
revision = '5d5e04ea01d5'
down_revision = 'd02c04effb34'


def upgrade():
    op.alter_column('sdn_journal', 'data',
                    existing_type=sa.PickleType(),
                    type_=sa.Text,
                    existing_nullable=True)
    op.alter_column('sdn_journal', 'state',
                    existing_type=sa.Enum(
                        sdn_const.PENDING, sdn_const.FAILED,
                        sdn_const.PROCESSING, sdn_const.COMPLETED,
                        name='state'),
                    type_=sa.Enum(
                        sdn_const.PENDING, sdn_const.FAILED,
                        sdn_const.PROCESSING, sdn_const.MONITORING,
                        sdn_const.COMPLETED,
                        name='state'),
                    existing_nullable=True)



# Copyright 2016 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from networking_mlnx._i18n import _
from networking_mlnx.plugins.ml2.drivers.sdn import constants as sdn_const

sdn_opts = [
        cfg.BoolOpt('sync_enabled',
                    help=_("Whether synchronising state to an SDN provider is "
                           "enabled."),
                    default=True),
        cfg.StrOpt('url',
                   help=_("HTTP URL of SDN Provider."),
                   ),
        cfg.StrOpt('domain',
                   help=_("Cloud domain name in SDN provider "
                          "(for example: cloudx)"),
                   default='cloudx'
                   ),
        cfg.StrOpt('username',
                   help=_("HTTP username for authentication."),
                   ),
        cfg.StrOpt('password',
                   help=_("HTTP password for authentication."),
                   secret=True,
                   default='123456'
                   ),
        cfg.IntOpt('timeout',
                   help=_("HTTP timeout in seconds."),
                   default=10
                   ),
        cfg.IntOpt('sync_timeout', default=10,
                   help=_("Sync thread timeout in seconds.")),
        cfg.IntOpt('retry_count', default=-1,
                   help=_("Number of times to retry a row "
                          "before failing."
                          "To disable retry count value should be -1")),
        cfg.IntOpt('maintenance_interval', default=300,
                   help=_("Journal maintenance operations interval "
                          "in seconds.")),
        cfg.IntOpt('completed_rows_retention', default=600,
                   help=_("Time to keep completed rows in seconds."
                          "Completed rows retention will be checked every "
                          "maintenance_interval by the cleanup thread."
                          "To disable completed rows deletion "
                          "value should be -1")),
        cfg.IntOpt('processing_timeout', default='100',
                   help=_("Time in seconds to wait before a "
                          "processing row is marked back to pending.")),
        cfg.ListOpt('physical_networks',
                default=sdn_const.ANY,
                help=_("Comma-separated list of <physical_network> "
                       "that it will send notification. * "
                       "means all physical_networks")),
        cfg.BoolOpt('bind_normal_ports',
                    default=False,
                    help=_("Allow the binding of normal ports for ports "
                           "associated with an InfiniBand physnet from "
                           "bind_normal_ports_physnets.")),
        # TODO(adrianc): The name here is a bit missleading, since there is no
        # usecase where SDN mechanism driver should bind normal ports for ETH
        # physnets. this should be renamed to: `infiniband_physnets` which will
        # require updates to deployment projects as well.
        cfg.ListOpt('bind_normal_ports_physnets',
                    default=[],
                    help=_("A list of InfiniBand physnets in which binding of "
                           "normal ports is allowed. This option is used in "
                           "conjuction with bind_normal_ports. "
                           "The list must be a subset of physical_networks")),
]



# Copyright 2015 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import datetime

import mock
import requests

from neutron.plugins.ml2 import plugin
from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron.tests.unit import testlib_api
from neutron_lib import constants
from neutron_lib import context as nl_context
from oslo_config import cfg
from oslo_config import fixture as fixture_config
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from networking_mlnx.db import db
from networking_mlnx.journal import cleanup
from networking_mlnx.journal import journal
from networking_mlnx.plugins.ml2.drivers.sdn import client
from networking_mlnx.plugins.ml2.drivers.sdn import config
from networking_mlnx.plugins.ml2.drivers.sdn import constants as sdn_const
from networking_mlnx.plugins.ml2.drivers.sdn import sdn_mech_driver
from networking_mlnx.plugins.ml2.drivers.sdn import utils as sdn_utils

PLUGIN_NAME = 'neutron.plugins.ml2.plugin.Ml2Plugin'
SEG_ID = 4
DEVICE_OWNER_COMPUTE = 'compute:None'
MECHANISM_DRIVER_NAME = 'mlnx_sdn_assist'


cfg.CONF.import_group(sdn_const.GROUP_OPT,
                      'networking_mlnx.plugins.ml2.drivers.sdn')


class SdnConfigBase(test_plugin.Ml2PluginV2TestCase):

    def setUp(self):
        super(SdnConfigBase, self).setUp()
        self.conf_fixture = self.useFixture(fixture_config.Config())
        self.conf = self.conf_fixture.conf
        self.conf.register_opts(config.sdn_opts, sdn_const.GROUP_OPT)
        self.conf.set_override('mechanism_drivers',
                              ['logger', MECHANISM_DRIVER_NAME],
                              'ml2')
        self.conf.set_override('url', 'http://127.0.0.1/neo',
                               sdn_const.GROUP_OPT)
        self.conf.set_override('username', 'admin', sdn_const.GROUP_OPT)


class SdnTestCase(SdnConfigBase):

    def setUp(self):
        super(SdnTestCase, self).setUp()
        self.mech = sdn_mech_driver.SDNMechanismDriver()
        mock.patch.object(journal.SdnJournalThread,
                          'start_sync_thread').start()
        self.mock_request = mock.patch.object(client.SdnRestClient,
                                              'request').start()
        self.mock_request.side_effect = self.check_request

    def check_request(self, method, urlpath, obj):
        self.assertFalse(urlpath.startswith("http://"))


class SdnMechanismConfigTests(testlib_api.SqlTestCase):

    def _set_config(self, url='http://127.0.0.1/neo',
                    username='admin',
                    password='123456',
                    sync_enabled=True):
        self.conf_fixture = self.useFixture(fixture_config.Config())
        self.conf = self.conf_fixture.conf
        self.conf.register_opts(config.sdn_opts, sdn_const.GROUP_OPT)
        self.conf.set_override('mechanism_drivers',
                               ['logger', MECHANISM_DRIVER_NAME],
                               'ml2')
        self.conf.set_override('url', url, sdn_const.GROUP_OPT)
        self.conf.set_override('username', username, sdn_const.GROUP_OPT)
        self.conf.set_override('password', password, sdn_const.GROUP_OPT)
        self.conf.set_override('sync_enabled', sync_enabled,
                               sdn_const.GROUP_OPT)

    def _test_missing_config(self, **kwargs):
        self._set_config(**kwargs)
        self.assertRaises(cfg.RequiredOptError,
                          plugin.Ml2Plugin)

    def test_valid_config(self):
        self._set_config()
        plugin.Ml2Plugin()

    def test_missing_url_raises_exception(self):
        self._test_missing_config(url=None)

    def test_missing_username_raises_exception(self):
        self._test_missing_config(username=None)

    def test_missing_password_raises_exception(self):
        self._test_missing_config(password=None)

    def test_missing_config_ok_when_disabled(self):
        self._set_config(url=None, username=None, password=None,
                         sync_enabled=False)
        plugin.Ml2Plugin()

    class SdnMechanismTestBasicGet(test_plugin.TestMl2BasicGet,
                                   SdnTestCase):
        pass

    class SdnMechanismTestNetworksV2(test_plugin.TestMl2NetworksV2,
                                     SdnTestCase):
        pass

    class SdnMechanismTestPortsV2(test_plugin.TestMl2PortsV2,
                                  SdnTestCase):
        pass


class DataMatcher(object):

    def __init__(self, context, object_type):
        self._data = context.__dict__["_" + object_type.lower()]

    def __eq__(self, data):
        return jsonutils.loads(data) == self._data

    def __repr__(self):
        return jsonutils.dumps(self._data)


class SdnDriverTestCase(SdnConfigBase):

    OPERATION_MAPPING = {
        sdn_const.PUT: 'update',
        sdn_const.DELETE: 'delete',
        sdn_const.POST: 'create',
    }

    def setUp(self):
        super(SdnDriverTestCase, self).setUp()
        context = nl_context.get_admin_context()
        self.db_session = context.session
        self.mech = sdn_mech_driver.SDNMechanismDriver()
        self.mock_sync_thread = mock.patch.object(
            journal.SdnJournalThread, 'start_sync_thread').start()
        self.mech.initialize()
        self.thread = journal.SdnJournalThread()
        self.addCleanup(self._db_cleanup)

    def _get_segments_list(self, seg_id=SEG_ID, net_type=constants.TYPE_VLAN):
        return [{'segmentation_id': seg_id,
                'physical_network': u'physnet1',
                'id': u'c13bba05-eb07-45ba-ace2-765706b2d701',
                'network_type': net_type}]

    def _get_mock_network_operation_context(self):
        current = {"provider:segmentation_id": SEG_ID,
                   'id': 'c13bba05-eb07-45ba-ace2-765706b2d701',
                   'name': 'net1',
                   'provider:network_type': 'vlan',
                   'network_qos_policy': None}
        context = mock.Mock(current=current, _network=current,
                            _segments=self._get_segments_list(),
                            network_segments=self._get_segments_list())
        context._plugin_context.session = (
            nl_context.get_admin_context().session)
        return context

    def _get_mock_port_operation_context(self):
        current = {'binding:host_id': 'r-ufm177',
                   'binding:profile': {u'pci_slot': u'0000:02:00.4',
                                       u'physical_network': u'physnet1',
                                       u'pci_vendor_info': u'15b3:1004'},
                   'id': '72c56c48-e9b8-4dcf-b3a7-0813bb3bd839',
                   'binding:vnic_type': 'direct',
                   'device_owner': DEVICE_OWNER_COMPUTE,
                   'mac_address': '12:34:56:78:21:b6',
                   'name': 'port_test1',
                   'network_id': 'c13bba05-eb07-45ba-ace2-765706b2d701',
                   'network_qos_policy': None}
        original = {'binding:host_id': None,
                    'binding:profile': {u'pci_slot': None,
                                        u'physical_network': u'physnet1',
                                        u'pci_vendor_info': u'15b3:1004'},
                    'id': None,
                    'binding:vnic_type': None,
                    'device_owner': DEVICE_OWNER_COMPUTE,
                    'mac_address': None,
                    'name': None,
                    'network_id': 'c13bba05-eb07-45ba-ace2-765706b2d701',
                    'network_qos_policy': None}

        # The port context should have NetwrokContext object that contain
        # the segments list
        network_context = type('NetworkContext', (object,),
                            {"network_segments": self._get_segments_list()})

        context = mock.Mock(current=current, _port=current,
                            original=original,
                            network=network_context)
        context._plugin_context.session = (
            nl_context.get_admin_context().session)
        return context

    def _get_mock_bind_operation_context(self):
        current = {'binding:host_id': 'r-ufm177',
                   'binding:profile': {u'pci_slot': u'0000:02:00.4',
                                       u'physical_network': u'physnet1',
                                       u'pci_vendor_info': u'15b3:1004'},
                   'id': '72c56c48-e9b8-4dcf-b3a7-0813bb3bd839',
                   'binding:vnic_type': 'direct',
                   'mac_address': '12:34:56:78:21:b6',
                   'name': 'port_test1',
                   'device_owner': DEVICE_OWNER_COMPUTE,
                   'network_id': 'c13bba05-eb07-45ba-ace2-765706b2d701',
                   'network_qos_policy': None}
        # The port context should have NetwrokContext object that contain
        # the segments list
        network_context = type('NetworkContext', (object,),
                            {"network_segments": self._get_segments_list()})
        context = mock.Mock(current=current, _port=current,
                            segments_to_bind=self._get_segments_list(),
                            network=network_context)
        context._plugin_context.session = (
            nl_context.get_admin_context().session)
        return context

    def _get_mock_operation_context(self, object_type):
        getter = getattr(self, '_get_mock_%s_operation_context' %
                         object_type.lower())
        return getter()

    _status_code_msgs = {
        200: '',
        201: '',
        204: '',
        400: '400 Client Error: Bad Request',
        401: '401 Client Error: Unauthorized',
        403: '403 Client Error: Forbidden',
        404: '404 Client Error: Not Found',
        409: '409 Client Error: Conflict',
        501: '501 Server Error: Not Implemented',
        503: '503 Server Error: Service Unavailable',
    }

    def _get_http_request_codes(self):
        for err_code in (requests.codes.ok,
                         requests.codes.created,
                         requests.codes.no_content,
                         requests.codes.bad_request,
                         requests.codes.unauthorized,
                         requests.codes.forbidden,
                         requests.codes.not_found,
                         requests.codes.conflict,
                         requests.codes.not_implemented,
                         requests.codes.service_unavailable):
            yield err_code

    def _db_cleanup(self):
        rows = db.get_all_db_rows(self.db_session)
        for row in rows:
            db.delete_row(self.db_session, row=row)

    @classmethod
    def _get_mock_request_response(cls, status_code, job_url):
        response = mock.Mock(status_code=status_code)
        if status_code < 400:
            response.raise_for_status = mock.Mock()
            response.json = mock.Mock(
                side_effect=[job_url, {"Status": "Completed"}])
        else:
            mock.Mock(side_effect=requests.exceptions.HTTPError(
                cls._status_code_msgs[status_code]))

        return response

    def _test_operation(self, method, status_code, expected_calls,
                        *args, **kwargs):
        job_url = 'app/jobs/' + uuidutils.generate_uuid()
        urlpath = sdn_utils.strings_to_url(
            cfg.CONF.sdn.url, job_url)

        request_response = self._get_mock_request_response(
            status_code, job_url)
        if expected_calls == 4 and status_code < 400:
            job_url2 = 'app/jobs/' + uuidutils.generate_uuid()
            urlpath2 = sdn_utils.strings_to_url(
                cfg.CONF.sdn.url, job_url)
            request_response.json = mock.Mock(
                side_effect=[job_url, job_url2,
                {"Status": "Completed"}, {"Status": "Completed"}])
        with mock.patch('requests.Session.request',
                        return_value=request_response) as mock_method:

            method(exit_after_run=True)
            login_args = mock.call(
                sdn_const.POST, mock.ANY,
                headers=sdn_const.LOGIN_HTTP_HEADER,
                data=mock.ANY, timeout=cfg.CONF.sdn.timeout)
            job_get_args = mock.call(
                sdn_const.GET, data=None,
                headers=sdn_const.JSON_HTTP_HEADER,
                url=urlpath, timeout=cfg.CONF.sdn.timeout)
            if status_code < 400:
                if expected_calls:
                    operation_args = mock.call(
                        headers=sdn_const.JSON_HTTP_HEADER,
                        timeout=cfg.CONF.sdn.timeout, *args, **kwargs)
                    if expected_calls == 4:
                        urlpath2 = sdn_utils.strings_to_url(
                            cfg.CONF.sdn.url, job_url2)
                        job_get_args2 = mock.call(
                            sdn_const.GET, data=None,
                            headers=sdn_const.JSON_HTTP_HEADER,
                            url=urlpath2, timeout=cfg.CONF.sdn.timeout)
                        self.assertEqual(
                            login_args, mock_method.mock_calls[4])
                        self.assertEqual(
                            job_get_args, mock_method.mock_calls[5])
                        self.assertEqual(
                            login_args, mock_method.mock_calls[6])
                        self.assertEqual(
                            job_get_args2, mock_method.mock_calls[7])
                    else:
                        self.assertEqual(
                            login_args, mock_method.mock_calls[0])
                        self.assertEqual(
                            operation_args, mock_method.mock_calls[1])
                        self.assertEqual(
                            login_args, mock_method.mock_calls[2])
                        self.assertEqual(
                            job_get_args, mock_method.mock_calls[3])

                # we need to reduce the login call_cout
                self.assertEqual(expected_calls * 2, mock_method.call_count)

    def _call_operation_object(self, operation, object_type):
        if object_type == sdn_const.PORT and operation == sdn_const.POST:
            context = self._get_mock_bind_operation_context()
            method = getattr(self.mech, 'bind_port')
        else:
            context = self._get_mock_operation_context(object_type)
            operation = self.OPERATION_MAPPING[operation]
            object_type = object_type.lower()
            method = getattr(self.mech, '%s_%s_precommit' % (operation,
                                                             object_type))
        method(context)

    def _test_operation_object(self, operation, object_type):
        self._call_operation_object(operation, object_type)

        context = self._get_mock_operation_context(object_type)
        row = db.get_oldest_pending_db_row_with_lock(self.db_session)
        self.assertEqual(operation, row['operation'])
        self.assertEqual(object_type, row['object_type'])
        self.assertEqual(context.current['id'], row['object_uuid'])

    def _test_thread_processing(self, operation, object_type,
                                expected_calls=2):
        status_codes = {sdn_const.POST: requests.codes.created,
                        sdn_const.PUT: requests.codes.ok,
                        sdn_const.DELETE: requests.codes.no_content}

        http_request = operation
        status_code = status_codes[operation]

        self._call_operation_object(operation, object_type)

        if object_type == sdn_const.PORT and operation == sdn_const.POST:
            context = self._get_mock_bind_operation_context()
        else:
            context = self._get_mock_operation_context(object_type)

        url_object_type = object_type.replace('_', '-')
        url = '%s/%s/%s' % (cfg.CONF.sdn.url,
                            cfg.CONF.sdn.domain,
                            url_object_type)
        if operation in (sdn_const.PUT, sdn_const.DELETE):
            uuid = context.current['id']
            url = '%s/%s' % (url, uuid)
        kwargs = {'url': url, 'data': DataMatcher(context, object_type)}
        with mock.patch.object(self.thread.event, 'wait',
                               return_value=False):
            self._test_operation(self.thread.run_sync_thread, status_code,
                                 expected_calls, http_request, **kwargs)

    def _test_object_type(self, object_type):
        # Add and process create request.
        self._test_thread_processing(sdn_const.POST, object_type)
        rows = db.get_all_db_rows_by_state(self.db_session,
                                           sdn_const.COMPLETED)
        self.assertEqual(1, len(rows))

        # Add and process update request. Adds to database.
        self._test_thread_processing(sdn_const.PUT, object_type)
        rows = db.get_all_db_rows_by_state(self.db_session,
                                           sdn_const.COMPLETED)
        self.assertEqual(2, len(rows))

        # Add and process update request. Adds to database.
        self._test_thread_processing(sdn_const.DELETE, object_type)
        rows = db.get_all_db_rows_by_state(self.db_session,
                                           sdn_const.COMPLETED)
        self.assertEqual(3, len(rows))

    def _test_object_type_pending_network(self, object_type):
        # Create a network (creates db row in pending state).
        self._call_operation_object(sdn_const.POST,
                                    sdn_const.NETWORK)

        # Create object_type database row and process. This results in both
        # the object_type and network rows being processed.
        self._test_thread_processing(sdn_const.POST, object_type,
                                     expected_calls=4)

        # Verify both rows are now marked as completed.
        rows = db.get_all_db_rows_by_state(self.db_session,
                                           sdn_const.COMPLETED)
        self.assertEqual(2, len(rows))

    def _test_object_type_processing_network(self, object_type):
        self._test_object_operation_pending_another_object_operation(
            object_type, sdn_const.POST, sdn_const.NETWORK,
            sdn_const.POST)

    def _test_object_operation_pending_object_operation(
        self, object_type, operation, pending_operation):
        self._test_object_operation_pending_another_object_operation(
            object_type, operation, object_type, pending_operation)

    def _test_object_operation_pending_another_object_operation(
        self, object_type, operation, pending_type, pending_operation):
        # Create the object_type (creates db row in pending state).
        self._call_operation_object(pending_operation,
                                    pending_type)

        # Get pending row and mark as processing so that
        # this row will not be processed by journal thread.
        row = db.get_all_db_rows_by_state(self.db_session, sdn_const.PENDING)
        db.update_db_row_state(self.db_session, row[0], sdn_const.PROCESSING)

        # Create the object_type database row and process.
        # Verify that object request is not processed because the
        # dependent object operation has not been marked as 'completed'.
        self._test_thread_processing(operation,
                                     object_type,
                                     expected_calls=0)

        # Verify that all rows are still in the database.
        rows = db.get_all_db_rows_by_state(self.db_session,
                                           sdn_const.PROCESSING)
        self.assertEqual(1, len(rows))
        rows = db.get_all_db_rows_by_state(self.db_session, sdn_const.PENDING)
        self.assertEqual(1, len(rows))

    def _test_parent_delete_pending_child_delete(self, parent, child):
        self._test_object_operation_pending_another_object_operation(
            parent, sdn_const.DELETE, child, sdn_const.DELETE)

    def _test_cleanup_processing_rows(self, last_retried, expected_state):
        # Create a dummy network (creates db row in pending state).
        self._call_operation_object(sdn_const.POST,
                                    sdn_const.NETWORK)

        # Get pending row and mark as processing and update
        # the last_retried time
        row = db.get_all_db_rows_by_state(self.db_session,
                                          sdn_const.PENDING)[0]
        row.last_retried = last_retried
        db.update_db_row_state(self.db_session, row, sdn_const.PROCESSING)

        # Test if the cleanup marks this in the desired state
        # based on the last_retried timestamp
        cleanup.JournalCleanup().cleanup_processing_rows(self.db_session)

        # Verify that the Db row is in the desired state
        rows = db.get_all_db_rows_by_state(self.db_session, expected_state)
        self.assertEqual(1, len(rows))

    def test_driver(self):
        for operation in (sdn_const.POST, sdn_const.PUT,
                          sdn_const.DELETE):
            for object_type in (sdn_const.NETWORK, sdn_const.PORT):
                self._test_operation_object(operation, object_type)

    def test_network(self):
        self._test_object_type(sdn_const.NETWORK)

    def test_network_update_pending_network_create(self):
        self._test_object_operation_pending_object_operation(
            sdn_const.NETWORK, sdn_const.PUT, sdn_const.POST)

    def test_network_delete_pending_network_create(self):
        self._test_object_operation_pending_object_operation(
            sdn_const.NETWORK, sdn_const.DELETE, sdn_const.POST)

    def test_network_delete_pending_network_update(self):
        self._test_object_operation_pending_object_operation(
            sdn_const.NETWORK, sdn_const.DELETE, sdn_const.PUT)

    def test_network_delete_pending_port_delete(self):
        self._test_parent_delete_pending_child_delete(
            sdn_const.NETWORK, sdn_const.PORT)

    def test_port(self):
        self._test_object_type(sdn_const.PORT)

    def test_port_update_pending_port_create(self):
        self._test_object_operation_pending_object_operation(
            sdn_const.PORT, sdn_const.PUT, sdn_const.POST)

    def test_port_delete_pending_port_create(self):
        self._test_object_operation_pending_object_operation(
            sdn_const.PORT, sdn_const.DELETE, sdn_const.POST)

    def test_port_delete_pending_port_update(self):
        self._test_object_operation_pending_object_operation(
            sdn_const.PORT, sdn_const.DELETE, sdn_const.PUT)

    def test_port_pending_network(self):
        self._test_object_type_pending_network(sdn_const.PORT)

    def test_port_processing_network(self):
        self._test_object_type_processing_network(sdn_const.PORT)

    def test_cleanup_processing_rows_time_not_expired(self):
        self._test_cleanup_processing_rows(datetime.datetime.utcnow(),
                                           sdn_const.PROCESSING)

    def test_cleanup_processing_rows_time_expired(self):
        old_time = datetime.datetime.utcnow() - datetime.timedelta(hours=24)
        self._test_cleanup_processing_rows(old_time, sdn_const.PENDING)

    def test_thread_call(self):
        """Verify that the sync thread method is called."""

        # Create any object that would spin up the sync thread via the
        # decorator call_thread_on_end() used by all the event handlers.
        self._call_operation_object(sdn_const.POST,
                                    sdn_const.NETWORK)

        # Verify that the thread call was made.
        self.assertTrue(self.mock_sync_thread.called)

    def _decrease_row_created_time(self, row):
        row.created_at -= datetime.timedelta(hours=1)
        self.db_session.merge(row)
        self.db_session.flush()

    def test_sync_multiple_updates(self):
        # add 2 updates
        for i in range(2):
            self._call_operation_object(sdn_const.PUT,
                                        sdn_const.NETWORK)

        # get the last update row
        last_row = db.get_all_db_rows(self.db_session)[-1]

        # change the last update created time
        self._decrease_row_created_time(last_row)

        # create 1 more operation to trigger the sync thread
        # verify that there are no calls to NEO controller, because the
        # first row was not valid (exit_after_run = true)
        self._test_thread_processing(sdn_const.PUT,
                                     sdn_const.NETWORK, expected_calls=0)

        # validate that all the rows are in 'pending' state
        # first row should be set back to 'pending' because it was not valid
        rows = db.get_all_db_rows_by_state(self.db_session, sdn_const.PENDING)
        self.assertEqual(3, len(rows))

    def test_network_filter_phynset(self):
        self.conf.set_override(
            'physical_networks', 'datacenter', sdn_const.GROUP_OPT)
        self.mech = sdn_mech_driver.SDNMechanismDriver()
        self.mech.initialize()
        self._test_filtered_object_type(sdn_const.NETWORK)

    def test_port_filter_phynset(self):
        self.conf.set_override(
            'physical_networks', 'datacenter', sdn_const.GROUP_OPT)
        self.mech = sdn_mech_driver.SDNMechanismDriver()
        self.mech.initialize()
        self._test_filtered_object_type(sdn_const.PORT)

    def _test_filtered_object_type(self, object_type):
        # Add and process create request.
        for operation in (sdn_const.POST, sdn_const.PUT, sdn_const.DELETE):
            self._call_operation_object(operation, object_type)
            rows = db.get_all_db_rows(self.db_session)
            self.assertEqual(0, len(rows))



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import neutron_lib.api.definitions.provider_net as prov_net_attr

SEGMENTATION_ID = prov_net_attr.SEGMENTATION_ID
PHYSICAL_NETWORK = prov_net_attr.PHYSICAL_NETWORK
ADMIN_STATE_UP = 'admin_state_up'

# Types of interfaces created by various neutron interface drivers
INTERFACE_KIND_UNKNOWN = 'unknown'
INTERFACE_KIND_OVS = 'openvswitch'
INTERFACE_KIND_VETH = 'veth'
INTERFACE_KIND_NULL = 'null'
INTERFACE_KIND_IPOIB = 'ipoib'



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from networking_mlnx.eswitchd.common import constants

LOG = logging.getLogger(__name__)


class eSwitchDB(object):

    def __init__(self, pf, vfs):
        self.port_table = {}
        self.port_policy = {}
        self.vfs = vfs
        self.pf = pf

    def create_port(self, port_name, port_type):
        self.port_table.update({port_name: {'type': port_type,
                                            'vnic': None,
                                            'state': None,
                                            'alias': None,
                                            'device_id': None}})

    def plug_nic(self, port_name):
        self.port_table[port_name]['state'] = constants.VPORT_STATE_ATTACHED
        LOG.info("port table:", self.port_table)

    def get_port_state(self, dev):
        state = None
        dev = self.port_table.get(dev)
        if dev:
            state = dev.get('state')
        return state

    def get_attached_vnics(self):
        vnics = {}
        for port in self.port_table.values():
            vnic_mac = port['vnic']
            state = port['state']
            if vnic_mac and state == constants.VPORT_STATE_ATTACHED:
                device_id = self.port_policy[vnic_mac]['device_id']
                vnics[vnic_mac] = {'mac': vnic_mac, 'device_id': device_id}
        return vnics

    def get_port_policy_matrix(self):
        table_matrix = [['VNIC_MAC', 'VLAN', 'DEV', 'DEVICE_ID']]
        for vnic_mac, port_policy in self.port_policy.items():
            table_matrix.append([vnic_mac, port_policy['vlan'],
                                 port_policy['dev'], port_policy['device_id']])
        return table_matrix

    def get_port_table(self):
        return self.port_table

    def get_port_table_matrix(self):
        table_matrix = [['PORT_NAME', 'TYPE', 'VNIC', 'STATE', 'ALIAS',
                         'DEVICE_ID']]
        for port_name, port_data in self.port_table.items():
            table_matrix.append([port_name, port_data['type'],
                                 port_data['vnic'],
                                 port_data['state'], port_data['alias'],
                                 port_data['device_id']])
        return table_matrix

    def create_vnic(self, vnic_mac):
        if not self.vnic_exists(vnic_mac):
            self.port_policy.update({vnic_mac: {'vlan': None,
                                                'dev': None,
                                                'device_id': None,
                                                'port_id': None}})

    def get_dev_for_vnic(self, vnic_mac):
        dev = None
        if vnic_mac in self.port_policy:
            if 'dev' in self.port_policy[vnic_mac]:
                dev = self.port_policy[vnic_mac]['dev']
        return dev

    def vnic_exists(self, vnic_mac):
        if vnic_mac in self.port_policy:
            return True
        else:
            return False

    def attach_vnic(self, port_name, device_id, vnic_mac, dev_name=None):
        self.port_table[port_name]['vnic'] = vnic_mac
        self.port_table[port_name]['alias'] = dev_name
        self.port_table[port_name]['state'] = constants.VPORT_STATE_PENDING
        self.port_table[port_name]['device_id'] = device_id
        dev = self.get_dev_for_vnic(vnic_mac)
        if not dev and vnic_mac != constants.INVALID_MAC:
            if vnic_mac in self.port_policy:
                vnic_mac_entry = self.port_policy[vnic_mac]
                vnic_mac_entry['dev'] = port_name
                vnic_mac_entry['device_id'] = device_id
                vnic_mac_entry.setdefault('vlan', None)
            else:
                self.port_policy.update({vnic_mac: {'vlan': None,
                                                    'dev': port_name,
                                                    'device_id': device_id,
                                                    }})
            return True
        return False

    def detach_vnic(self, vnic_mac):
        dev = self.get_dev_for_vnic(vnic_mac)
        if dev:
            self.port_table[dev]['vnic'] = None
            self.port_table[dev]['alias'] = None
            self.port_table[dev]['state'] = constants.VPORT_STATE_UNPLUGGED
            self.port_table[dev]['device_id'] = None
        return dev

    def port_release(self, vnic_mac):
        try:
            dev = self.get_dev_for_vnic(vnic_mac)
            vnic = self.port_policy.pop(vnic_mac)
            self.port_table[dev]['state'] = None
            vnic['type'] = self.port_table[vnic['dev']]['type']
            return vnic
        except KeyError:
            return
        except IndexError:
            return

    def set_vlan(self, vnic_mac, vlan):
        if not self.vnic_exists(vnic_mac):
            self.create_vnic(vnic_mac)

        self.port_policy[vnic_mac]['vlan'] = vlan



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


def get_root_helper():
    root_helper = 'sudo neutron-rootwrap %s' % cfg.CONF.DAEMON.rootwrap_conf
    return root_helper


def execute(*cmd, **kwargs):
    if kwargs.get('root_helper') is None:
        kwargs['run_as_root'] = True
        kwargs['root_helper'] = get_root_helper()
    return processutils.execute(*cmd, **kwargs)



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock
from oslo_concurrency import lockutils
from oslo_utils.fixture import uuidsentinel as uuids

from networking_mlnx.linux.interface_drivers import constants
from networking_mlnx.linux.interface_drivers import network_cache
from networking_mlnx.tests import base

network_db = {
    uuids.obj_a: {'id': uuids.obj_a, 'attrA': 'a-attr', 'attrB': 'b-attr'},
    uuids.obj_b: {'id': uuids.obj_b, 'attrA': 'a-attr', 'attrB': 'b-attr'},
    uuids.obj_c: {'id': uuids.obj_c, 'attrA': 'a-attr', 'attrB': 'b-attr'},
}


class TestSimpleCache(base.TestCase):
    def setUp(self):
        super(TestSimpleCache, self).setUp()
        self.cache = network_cache.SimpleCache()
        self._init_cache()

    def _init_cache(self):
        self.cache.object_cache = copy.deepcopy(network_db)

    def test_contains_operator(self):
        self.assertIn(uuids.obj_a, self.cache)
        self.assertNotIn(uuids.obj_non_existent, self.cache)
        self.assertNotIn(uuids.obj_non_existent, self.cache.object_cache)

    def test_get(self):
        obj = self.cache.get(uuids.obj_b)
        self.assertDictEqual(network_db[uuids.obj_b], obj)

    def test_get_non_existent_entry(self):
        self.assertIsNone(self.cache.get(uuids.obj_non_existent))

    def test_get_all(self):
        objs = self.cache.get_all()
        objs_ids = set([obj['id'] for obj in objs])
        self.assertEqual(objs_ids, set(network_db.keys()))
        for obj in objs:
            self.assertDictEqual(network_db[obj['id']], obj)

    def test_put(self):
        new_entry = {'id': uuids.new_obj, 'attrA': 'a-attr', 'attrB': 'b-attr'}
        self.cache.put(uuids.new_obj, new_entry)
        obj = self.cache.get(uuids.new_obj)
        self.assertDictEqual(new_entry, obj)

    def test_remove(self):
        self.cache.remove(uuids.obj_a)
        self.assertIsNone(self.cache.get(uuids.obj_a))
        self.assertNotIn(uuids.obj_a, self.cache.object_cache)

    def test_clear(self):
        self.cache.clear()
        self.assertListEqual([], self.cache.get_all())


class TestNetworkCache(base.TestCase):
    def _get_networks_cb(self, filters, fields):
        if 'id' in filters:
            net_id = filters["id"][0]
            nets = [self.network_db[net_id]]
        else:
            nets = list(self.network_db.values())
        return nets

    def setUp(self):
        super(TestNetworkCache, self).setUp()
        self.cb_mock = mock.Mock()
        self.cb_mock.side_effect = self._get_networks_cb
        self.network_fields = ['attrA', 'attrB', 'id']
        self.cache = network_cache.NetworkCache(
            self.cb_mock, self.network_fields)
        self.network_db = copy.deepcopy(network_db)

    def test_get(self):
        self.cache.put(uuids.obj_a, self.network_db[uuids.obj_a])
        obj = self.cache.get(uuids.obj_a)
        self.assertDictEqual(self.network_db[uuids.obj_a], obj)

    def test_get_with_miss(self):
        self.cache.put(uuids.obj_a, self.network_db[uuids.obj_a])
        obj = self.cache.get(uuids.obj_b)
        self.assertTrue(self.cb_mock.called)
        self.assertDictEqual(self.network_db[uuids.obj_b], obj)

    def test_refresh(self):
        self.cache.refresh()
        self.cb_mock.assert_called_once_with(
            filters={constants.ADMIN_STATE_UP: [True]},
            fields=self.network_fields)
        objs = self.cache.get_all()
        self.assertEqual(
            set(self.network_db.keys()), set([obj['id'] for obj in objs]))

    def test_remove_stale_networks(self):
        self.cache.put(uuids.obj_a, self.network_db[uuids.obj_a])
        self.cache.put(uuids.obj_b, self.network_db[uuids.obj_b])
        del self.network_db[uuids.obj_a]
        self.cache.remove_stale_networks()
        self.assertNotIn(uuids.obj_a, self.cache)
        self.assertIn(uuids.obj_b, self.cache)


class TestSafeNetworkCache(TestNetworkCache):
    def setUp(self):
        super(TestSafeNetworkCache, self).setUp()
        with mock.patch.object(lockutils, 'ReaderWriterLock') as rw_cls:
            self.rw_lock_mock = rw_cls.return_value = mock.MagicMock()
            self.cache = network_cache.SafeNetworkCache(
                self.cb_mock, ['attrA', 'attrB'])

    def test_get(self):
        self.cache.get(uuids.obj_a)
        self.assertTrue(self.rw_lock_mock.read_lock.called)

    def test_get_all(self):
        self.cache.get_all()
        self.assertTrue(self.rw_lock_mock.read_lock.called)

    def test_put(self):
        self.cache.put(uuids.new_obj_id, {})
        self.assertTrue(self.rw_lock_mock.write_lock.called)

    def test_remove(self):
        self.cache.remove(uuids.obj_a)
        self.assertTrue(self.rw_lock_mock.write_lock.called)

    def test_clear(self):
        self.cache.clear()
        self.assertTrue(self.rw_lock_mock.write_lock.called)

    def test_refresh(self):
        self.cache.refresh()
        self.assertTrue(self.rw_lock_mock.write_lock.called)

    def test_cache_miss_with_real_lock(self):
        # re-instantiate cache with a real lock
        self.cache = self.cache = network_cache.SafeNetworkCache(
                self.cb_mock, ['attrA', 'attrB'])
        self.test_get_with_miss()



# Copyright 2016 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from datetime import timedelta

import mock
from neutron.tests.unit import testlib_api
from neutron_lib import context
from oslo_db import exception

from networking_mlnx.db import db
from networking_mlnx.db.models import sdn_journal_db
from networking_mlnx.db.models import sdn_maintenance_db
from networking_mlnx.plugins.ml2.drivers.sdn import constants as sdn_const


class DbTestCase(testlib_api.SqlTestCaseLight):

    UPDATE_ROW = [sdn_const.NETWORK, 'id', sdn_const.PUT,
                  {'test': 'data'}]

    def setUp(self):
        super(DbTestCase, self).setUp()
        self.db_context = context.get_admin_context()
        self.db_session = self.db_context.session
        self.addCleanup(self._db_cleanup)

    def _db_cleanup(self):
        self.db_session.query(sdn_journal_db.SdnJournal).delete()

    def _update_row(self, row):
        self.db_session.merge(row)
        self.db_session.flush()

    def _test_validate_updates(self, rows, time_deltas, expected_validations):
        for row in rows:
            db.create_pending_row(self.db_session, *row)

        # update row created_at
        rows = db.get_all_db_rows(self.db_session)
        now = datetime.now()
        for row, time_delta in zip(rows, time_deltas):
            row.created_at = now - timedelta(hours=time_delta)
            self._update_row(row)

        # validate if there are older rows
        for row, expected_valid in zip(rows, expected_validations):
            valid = not db.check_for_older_ops(self.db_session, row)
            self.assertEqual(expected_valid, valid)

    def _test_retry_count(self, retry_num, max_retry,
                          expected_retry_count, expected_state):
        # add new pending row
        db.create_pending_row(self.db_session, *self.UPDATE_ROW)

        # update the row with the requested retry_num
        row = db.get_all_db_rows(self.db_session)[0]
        row.retry_count = retry_num - 1
        db.update_pending_db_row_retry(self.db_session, row, max_retry)

        # validate the state and the retry_count of the row
        row = db.get_all_db_rows(self.db_session)[0]
        self.assertEqual(expected_state, row.state)
        self.assertEqual(expected_retry_count, row.retry_count)

    def _test_update_row_state(self, from_state, to_state):
        # add new pending row
        db.create_pending_row(self.db_session, *self.UPDATE_ROW)

        row = db.get_all_db_rows(self.db_session)[0]
        for state in [from_state, to_state]:
            # update the row state
            db.update_db_row_state(self.db_session, row, state)

            # validate the new state
            row = db.get_all_db_rows(self.db_session)[0]
            self.assertEqual(state, row.state)

    def test_validate_updates_same_object_uuid(self):
        self._test_validate_updates(
            [self.UPDATE_ROW, self.UPDATE_ROW], [1, 0], [True, False])

    def test_validate_updates_same_created_time(self):
        self._test_validate_updates(
            [self.UPDATE_ROW, self.UPDATE_ROW], [0, 0], [True, True])

    def test_validate_updates_different_object_uuid(self):
        other_row = list(self.UPDATE_ROW)
        other_row[1] += 'a'
        self._test_validate_updates(
            [self.UPDATE_ROW, other_row], [1, 0], [True, True])

    def test_validate_updates_different_object_type(self):
        other_row = list(self.UPDATE_ROW)
        other_row[0] = sdn_const.PORT
        other_row[1] += 'a'
        self._test_validate_updates(
            [self.UPDATE_ROW, other_row], [1, 0], [True, True])

    def test_get_oldest_pending_row_none_when_no_rows(self):
        row = db.get_oldest_pending_db_row_with_lock(self.db_session)
        self.assertIsNone(row)

    def _test_get_oldest_pending_row_none(self, state):
        db.create_pending_row(self.db_session, *self.UPDATE_ROW)
        row = db.get_all_db_rows(self.db_session)[0]
        row.state = state
        self._update_row(row)

        row = db.get_oldest_pending_db_row_with_lock(self.db_session)
        self.assertIsNone(row)

    def test_get_oldest_pending_row_none_when_row_processing(self):
        self._test_get_oldest_pending_row_none(sdn_const.PROCESSING)

    def test_get_oldest_pending_row_none_when_row_failed(self):
        self._test_get_oldest_pending_row_none(sdn_const.FAILED)

    def test_get_oldest_pending_row_none_when_row_completed(self):
        self._test_get_oldest_pending_row_none(sdn_const.COMPLETED)

    def test_get_oldest_pending_row_none_when_row_monitoring(self):
        self._test_get_oldest_pending_row_none(sdn_const.MONITORING)

    def test_get_oldest_pending_row(self):
        db.create_pending_row(self.db_session, *self.UPDATE_ROW)
        row = db.get_oldest_pending_db_row_with_lock(self.db_session)
        self.assertIsNotNone(row)
        self.assertEqual(sdn_const.PROCESSING, row.state)

    def test_get_oldest_pending_row_order(self):
        db.create_pending_row(self.db_session, *self.UPDATE_ROW)
        older_row = db.get_all_db_rows(self.db_session)[0]
        older_row.last_retried -= timedelta(minutes=1)
        self._update_row(older_row)

        db.create_pending_row(self.db_session, *self.UPDATE_ROW)
        row = db.get_oldest_pending_db_row_with_lock(self.db_session)
        self.assertEqual(older_row, row)

    def test_get_all_monitoring_db_row_by_oldest_order(self):
        db.create_pending_row(self.db_session, *self.UPDATE_ROW)
        db.create_pending_row(self.db_session, *self.UPDATE_ROW)
        older_row = db.get_all_db_rows(self.db_session)[1]
        older_row.last_retried -= timedelta(minutes=1)
        older_row.state = sdn_const.MONITORING
        self._update_row(older_row)
        newer_row = db.get_all_db_rows(self.db_session)[0]
        newer_row.state = sdn_const.MONITORING
        self._update_row(newer_row)

        rows = db.get_all_monitoring_db_row_by_oldest(self.db_session)
        self.assertEqual(older_row, rows[0])
        self.assertEqual(newer_row, rows[1])

    def test_get_oldest_pending_row_when_deadlock(self):
        db.create_pending_row(self.db_session, *self.UPDATE_ROW)
        update_mock = (
            mock.MagicMock(side_effect=(exception.DBDeadlock, mock.DEFAULT)))

        # Mocking is mandatory to achieve a deadlock regardless of the DB
        # backend being used when running the tests
        with mock.patch.object(db, 'update_db_row_state', new=update_mock):
            row = db.get_oldest_pending_db_row_with_lock(self.db_session)
            self.assertIsNotNone(row)

        self.assertEqual(2, update_mock.call_count)

    def _test_delete_rows_by_state_and_time(self, last_retried, row_retention,
                                            state, expected_rows):
        db.create_pending_row(self.db_session, *self.UPDATE_ROW)

        # update state and last retried
        row = db.get_all_db_rows(self.db_session)[0]
        row.state = state
        row.last_retried = row.last_retried - timedelta(seconds=last_retried)
        self._update_row(row)

        db.delete_rows_by_state_and_time(self.db_session,
                                         sdn_const.COMPLETED,
                                         timedelta(seconds=row_retention))

        # validate the number of rows in the journal
        rows = db.get_all_db_rows(self.db_session)
        self.assertEqual(expected_rows, len(rows))

    def test_delete_completed_rows_no_new_rows(self):
        self._test_delete_rows_by_state_and_time(0, 10, sdn_const.COMPLETED, 1)

    def test_delete_completed_rows_one_new_row(self):
        self._test_delete_rows_by_state_and_time(6, 5, sdn_const.COMPLETED, 0)

    def test_delete_completed_rows_wrong_state(self):
        self._test_delete_rows_by_state_and_time(10, 8, sdn_const.PENDING, 1)

    def test_valid_retry_count(self):
        self._test_retry_count(1, 1, 1, sdn_const.PENDING)

    def test_invalid_retry_count(self):
        self._test_retry_count(2, 1, 1, sdn_const.FAILED)

    def test_update_row_state_to_pending(self):
        self._test_update_row_state(sdn_const.PROCESSING, sdn_const.PENDING)

    def test_update_row_state_to_processing(self):
        self._test_update_row_state(sdn_const.PENDING, sdn_const.PROCESSING)

    def test_update_row_state_to_failed(self):
        self._test_update_row_state(sdn_const.PROCESSING, sdn_const.FAILED)

    def test_update_row_state_to_monitoring(self):
        self._test_update_row_state(sdn_const.PROCESSING, sdn_const.MONITORING)

    def test_update_row_state_to_completed(self):
        self._test_update_row_state(sdn_const.PROCESSING, sdn_const.COMPLETED)

    def test_update_row_job_id(self):
        # add new pending row
        expected_job_id = 'job_id'
        db.create_pending_row(self.db_session, *self.UPDATE_ROW)
        row = db.get_all_db_rows(self.db_session)[0]
        db.update_db_row_job_id(self.db_session, row, expected_job_id)
        row = db.get_all_db_rows(self.db_session)[0]
        self.assertEqual(expected_job_id, row.job_id)

    def _test_maintenance_lock_unlock(self, db_func, existing_state,
                                      expected_state, expected_result):
        row = sdn_maintenance_db.SdnMaintenance(id='test',
                                             state=existing_state)
        self.db_session.add(row)
        self.db_session.flush()

        self.assertEqual(expected_result, db_func(self.db_session))
        row = self.db_session.query(sdn_maintenance_db.SdnMaintenance).one()
        self.assertEqual(expected_state, row['state'])

    def test_lock_maintenance(self):
        self._test_maintenance_lock_unlock(db.lock_maintenance,
                                           sdn_const.PENDING,
                                           sdn_const.PROCESSING,
                                           True)

    def test_lock_maintenance_fails_when_processing(self):
        self._test_maintenance_lock_unlock(db.lock_maintenance,
                                           sdn_const.PROCESSING,
                                           sdn_const.PROCESSING,
                                           False)

    def test_unlock_maintenance(self):
        self._test_maintenance_lock_unlock(db.unlock_maintenance,
                                           sdn_const.PROCESSING,
                                           sdn_const.PENDING,
                                           True)

    def test_unlock_maintenance_fails_when_pending(self):
        self._test_maintenance_lock_unlock(db.unlock_maintenance,
                                           sdn_const.PENDING,
                                           sdn_const.PENDING,
                                           False)



# Copyright 2014 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from oslo_config import cfg
from oslo_config import fixture as fixture_config

from networking_mlnx.plugins.ml2.drivers.sdn import client
from networking_mlnx.plugins.ml2.drivers.sdn import config
from networking_mlnx.plugins.ml2.drivers.sdn import constants as sdn_const
from networking_mlnx.tests import base


class TestClient(base.TestCase):

    def setUp(self):
        super(TestClient, self).setUp()
        self.conf_fixture = self.useFixture(fixture_config.Config())
        self.conf = self.conf_fixture.conf
        self._set_args()
        self.client = client.SdnRestClient.create_client()
        self.url = 'some_url'
        self.data = {'some': 'data'}

    def _set_args(self):
        self.conf.register_opts(config.sdn_opts, sdn_const.GROUP_OPT)
        for arg in client.SdnRestClient.MANDATORY_ARGS:
            self.conf.set_override(name=arg,
                                   override="http://some_val",
                                   group=sdn_const.GROUP_OPT)

    def test_mandatory_args(self):
        mandatory_arg_objects = filter(
                lambda obj: obj.name in client.SdnRestClient.MANDATORY_ARGS,
                config.sdn_opts)
        for arg in mandatory_arg_objects:
            self._set_args()
            self.conf.unregister_opt(opt=arg,
                                     group=sdn_const.GROUP_OPT)
            self.assertRaises(cfg.NoSuchOptError,
                              client.SdnRestClient.create_client)

    @mock.patch('networking_mlnx.plugins.ml2.drivers.'
                'sdn.client.SdnRestClient.request')
    def test_get(self, mocked_request):
        self.client.get(self.url, self.data)
        expected_url = '/'.join((self.conf.sdn.url, self.url))
        mocked_request.assert_called_once_with(sdn_const.GET,
                                               expected_url,
                                               self.data)

        mocked_request.reset_mock()
        self.client.get(self.url)
        mocked_request.assert_called_once_with(sdn_const.GET,
                                               expected_url,
                                               None)

        mocked_request.reset_mock()
        self.client.get()
        mocked_request.assert_called_once_with(sdn_const.GET,
                                               self.conf.sdn.url,
                                               None)

    @mock.patch('networking_mlnx.plugins.ml2.drivers.'
                'sdn.client.SdnRestClient.request')
    def test_put(self, mocked_request):
        self.client.put(self.url, self.data)
        expected_url = '/'.join((self.conf.sdn.url,
                                 self.conf.sdn.domain,
                                 self.url))
        mocked_request.assert_called_once_with(sdn_const.PUT,
                                               expected_url,
                                               self.data)

        mocked_request.reset_mock()
        self.client.put(self.url)
        mocked_request.assert_called_once_with(sdn_const.PUT,
                                               expected_url,
                                               None)

        mocked_request.reset_mock()
        self.client.put()
        expected_url = '/'.join((self.conf.sdn.url, self.conf.sdn.domain))
        mocked_request.assert_called_once_with(sdn_const.PUT,
                                               expected_url,
                                               None)

    @mock.patch('networking_mlnx.plugins.ml2.drivers.'
                'sdn.client.SdnRestClient.request')
    def test_post(self, mocked_request):
        self.client.post(self.url, self.data)
        expected_url = '/'.join((self.conf.sdn.url,
                                 self.conf.sdn.domain,
                                 self.url))
        mocked_request.assert_called_once_with(sdn_const.POST,
                                               expected_url,
                                               self.data)

        mocked_request.reset_mock()
        self.client.post(self.url)
        mocked_request.assert_called_once_with(sdn_const.POST,
                                               expected_url,
                                               None)

        mocked_request.reset_mock()
        self.client.post()
        expected_url = '/'.join((self.conf.sdn.url, self.conf.sdn.domain))
        mocked_request.assert_called_once_with(sdn_const.POST,
                                               expected_url,
                                               None)

    @mock.patch('networking_mlnx.plugins.ml2.drivers.'
                'sdn.client.SdnRestClient.request')
    def test_delete(self, mocked_request):
        self.client.delete(self.url, self.data)
        expected_url = '/'.join((self.conf.sdn.url,
                                 self.conf.sdn.domain,
                                 self.url))
        mocked_request.assert_called_once_with(sdn_const.DELETE,
                                               expected_url,
                                               self.data)

        mocked_request.reset_mock()
        self.client.delete(self.url)
        mocked_request.assert_called_once_with(sdn_const.DELETE,
                                               expected_url,
                                               None)

        mocked_request.reset_mock()
        self.client.delete()
        expected_url = '/'.join((self.conf.sdn.url, self.conf.sdn.domain))
        mocked_request.assert_called_once_with(sdn_const.DELETE,
                                               expected_url,
                                               None)

    @mock.patch('networking_mlnx.plugins.ml2.drivers.'
                'sdn.client.SdnRestClient._get_session',
                return_value=mock.Mock())
    def test_request_bad_data(self, mocked_get_session):
        # non serialized json data
        data = self
        self.assertRaises(ValueError,
                          self.client.request,
                          sdn_const.DELETE, '', data)



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import glob
import os
import re

from oslo_log import log as logging

from networking_mlnx._i18n import _LE
from networking_mlnx.eswitchd.common import constants

LOG = logging.getLogger(__name__)


class pciUtils(object):

    ETH_PATH = "/sys/class/net/%(interface)s"
    ETH_DEV = ETH_PATH + "/device"
    ETH_DRIVER = ETH_DEV + "/driver"
    ETH_PORT = ETH_PATH + "/dev_id"
    INFINIBAND_PATH = 'device/infiniband'
    VENDOR_PATH = ETH_DEV + '/vendor'
    _VIRTFN_RE = re.compile(r'virtfn(?P<vf_num>\d+)')
    VFS_PATH = ETH_DEV + "/virtfn*"

    def get_vfs_info(self, pf):
        vfs_info = {}
        try:
            dev_path = self.ETH_DEV % {'interface': pf}
            dev_info = os.listdir(dev_path)
            for dev_filename in dev_info:
                result = self._VIRTFN_RE.match(dev_filename)
                if result and result.group('vf_num'):
                    dev_file = os.path.join(dev_path, dev_filename)
                    vf_pci = os.readlink(dev_file).strip("./")
                    vf_num = result.group('vf_num')
                    vf_device_type = self.get_pf_device_type(pf)
                    vfs_info[vf_pci] = {'vf_num': vf_num,
                                        'vf_device_type': vf_device_type}
        except Exception:
            LOG.error("PCI device %s not found", pf)
        return vfs_info

    def get_dev_attr(self, attr_path):
        try:
            fd = open(attr_path)
            return fd.readline().strip()
        except IOError:
            return

    def verify_vendor_pf(self, pf, vendor_id=constants.VENDOR):
        vendor_path = pciUtils.VENDOR_PATH % {'interface': pf}
        if self.get_dev_attr(vendor_path) == vendor_id:
            return True
        else:
            return False

    def get_pf_device_type(self, pf):
        device_type = None
        try:
            driver_type = os.readlink(self.ETH_DRIVER
                                      % {'interface': pf})
            driver_type = os.path.basename(driver_type)
            if driver_type == constants.MLNX4_DRIVER_TYPE:
                device_type = constants.MLNX4_DEVICE_TYPE
            elif driver_type == constants.MLNX5_DRIVER_TYPE:
                device_type = constants.MLNX5_DEVICE_TYPE
            else:
                raise Exception(_LE('driver type %s is not supported'),
                                driver_type)
        except IOError:
            pass
        return device_type

    def is_sriov_pf(self, pf):
        vfs_path = pciUtils.VFS_PATH % {'interface': pf}
        vfs = glob.glob(vfs_path)
        if vfs:
            return True
        else:
            return

    def get_pf_mlx_dev(self, pf):
        dev_path = (
            os.path.join(pciUtils.ETH_PATH % {'interface': pf},
            pciUtils.INFINIBAND_PATH))
        dev_info = os.listdir(dev_path)
        return dev_info.pop()

    def get_guid_index(self, pf_mlx_dev, dev, hca_port):
        guid_index = None
        path = constants.MLNX4_GUID_INDEX_PATH % (pf_mlx_dev, dev, hca_port)
        with open(path) as fd:
            guid_index = fd.readline().strip()
        return guid_index

    def get_eth_port(self, dev):
        port_path = pciUtils.ETH_PORT % {'interface': dev}
        try:
            with open(port_path) as f:
                dev_id = int(f.read(), 0)
                return dev_id + 1
        except IOError:
            return

    def get_vfs_macs_ib(self, fabric_details):
        macs_map = {}
        for pf_fabric_details in fabric_details.values():
            if (pf_fabric_details['pf_device_type'] ==
                constants.MLNX4_DEVICE_TYPE):
                macs_map.update(self.get_vfs_macs_ib_mlnx4(pf_fabric_details))
            elif (pf_fabric_details['pf_device_type'] ==
                  constants.MLNX5_DEVICE_TYPE):
                macs_map.update(self.get_vfs_macs_ib_mlnx5(pf_fabric_details))
        return macs_map

    def get_vfs_macs_ib_mlnx4(self, fabric_details):
        hca_port = fabric_details['hca_port']
        pf_mlx_dev = fabric_details['pf_mlx_dev']
        macs_map = {}
        guids_path = constants.MLNX4_ADMIN_GUID_PATH % (pf_mlx_dev, hca_port,
                                                  '[1-9]*')
        paths = glob.glob(guids_path)
        for path in paths:
            vf_index = path.split('/')[-1]
            with open(path) as f:
                guid = f.readline().strip()
                if guid == constants.MLNX4_INVALID_GUID:
                    mac = constants.INVALID_MAC
                else:
                    head = guid[:6]
                    tail = guid[-6:]
                    mac = ":".join(re.findall('..?', head + tail))
                macs_map[str(int(vf_index))] = mac
        return macs_map

    def get_vfs_macs_ib_mlnx5(self, fabric_details):
        vfs = fabric_details['vfs']
        macs_map = {}
        for vf in vfs.values():
            vf_num = vf['vf_num']
            pf_mlx_dev = fabric_details['pf_mlx_dev']
            guid_path = (
                constants.MLNX5_GUID_NODE_PATH % {'module': pf_mlx_dev,
                                                  'vf_num': vf_num})
            with open(guid_path) as f:
                guid = f.readline().strip()
                head = guid[:8]
                tail = guid[-9:]
                mac = head + tail
            macs_map[vf_num] = mac
        return macs_map

    def get_device_address(self, hostdev):
        domain = hostdev.attrib['domain'][2:]
        bus = hostdev.attrib['bus'][2:]
        slot = hostdev.attrib['slot'][2:]
        function = hostdev.attrib['function'][2:]
        dev = "%.4s:%.2s:%2s.%.1s" % (domain, bus, slot, function)
        return dev



# Copyright 2019 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading

from neutron.agent.common import utils as n_agent_utils
from neutron.agent.linux import interface as n_interface
from neutron.agent.linux import ip_lib as n_ip_lib
from neutron_lib.utils import helpers
from neutron_lib.utils import runtime
from oslo_config import cfg
from oslo_log import log
from oslo_service import loopingcall
import six

from networking_mlnx.linux.interface_drivers import config
from networking_mlnx.linux.interface_drivers import constants
from networking_mlnx.linux.interface_drivers import network_cache
from networking_mlnx.linux import ip_lib

# Register interface driver Opts when module is loaded
cfg.CONF.register_opts(config.interface_driver_opts)
LOG = log.getLogger(__name__)

DEFAULT_CACHE_CLEANING_INTERVAL = 86400


class IPoIBInterfaceDriver(n_interface.LinuxInterfaceDriver):
    """Driver for creating ipoib interfaces."""

    def __init__(self, conf, get_networks_callback=None, **kwargs):
        super(IPoIBInterfaceDriver, self).__init__(
            conf, get_networks_callback=get_networks_callback, **kwargs)
        self.get_networks = get_networks_callback
        self.root_dev = conf.ipoib_physical_interface
        if not n_ip_lib.device_exists(self.root_dev):
            LOG.error("IPoIB root device %s does not exist.", self.root_dev)

    def plug_new(self, network_id, port_id, device_name, mac_address,
                 bridge=None, namespace=None, prefix=None, mtu=None):
        """Plugin the interface."""
        ip = ip_lib.IPoIBWrapper(namespace=namespace)
        try:
            net = self.get_networks(filters={"id": [network_id]},
                                    fields=[constants.SEGMENTATION_ID])[0]
            segmentation_id = net.get(constants.SEGMENTATION_ID)
            dev = ip.add_ipoib(device_name, self.root_dev, segmentation_id)
            dev.link.set_up()
        except RuntimeError as e:
            LOG.error("Failed plugging interface '%s' - %s",
                      device_name, str(e))

    def unplug(self, device_name, bridge=None, namespace=None, prefix=None):
        """Unplug the interface."""
        ip = ip_lib.IPoIBWrapper(namespace=namespace)
        try:
            ip.del_ipoib(device_name)
        except RuntimeError as e:
            LOG.error("Failed unplugging interface '%s' - %s",
                      device_name, str(e))


class MultiInterfaceDriver(n_interface.LinuxInterfaceDriver):
    """Driver for invoking interface driver per physnet.

    This driver Delegates the plug/unplug methods
    to other interface drivers per physnet
    according to <physnet: driver> mapping provided in conf file.

    This driver assumes uniqueness in regards to the kind of interface
    created by each delegated driver.
    """

    # Mapping between interface driver and the Type of interface it creates
    # during a call to 'plug()' method.

    # NOTE(adrianc): OVSInterfaceDriver can create either veth or openvswitch
    #                interfaces, however in the mapping we assume that the
    #                latter will be used. It is the responsibility of the
    #                user of this map to check `ovs_use_veth` is false.

    driver_to_interface_kind = {
        n_interface.BridgeInterfaceDriver: constants.INTERFACE_KIND_VETH,
        n_interface.OVSInterfaceDriver: constants.INTERFACE_KIND_OVS,
        n_interface.NullDriver: constants.INTERFACE_KIND_NULL,
        IPoIBInterfaceDriver: constants.INTERFACE_KIND_IPOIB}

    network_cache = None
    _cache_init_lock = threading.RLock()

    def __init__(self, conf, get_networks_callback=None, **kwargs):
        super(MultiInterfaceDriver, self).__init__(
            conf, get_networks_callback=get_networks_callback, **kwargs)
        fields = ['id', constants.PHYSICAL_NETWORK, constants.SEGMENTATION_ID]
        MultiInterfaceDriver._init_network_cache(conf, get_networks_callback,
                                                 fields)
        # Use a cache backed get_networks callback to avoid un-needed RPC calls
        self.drivers = MultiInterfaceDriver.load_interface_driver_mappings(
            conf,
            get_networks_callback=(
                MultiInterfaceDriver._get_networks_from_cache),
            **kwargs)

    @staticmethod
    def _check_drivers(driver_mapping):
        """Check the loaded interface drivers are dealing with a unique
        interface kind.

        :raises SystemExit of 1 in case an inconsistency was found.
        """
        uniq_intf_drivers = []
        driver_types = []
        for driver in six.itervalues(driver_mapping):
            if driver.__class__ not in driver_types:
                driver_types.append(driver.__class__)
                uniq_intf_drivers.append(driver)

        interface_kinds = [driver._interface_kind_ for
                           driver in uniq_intf_drivers]
        if len(interface_kinds) != len(set(interface_kinds)):
            LOG.error("MultiInterfaceDriver cannot operate with interface"
                      "drivers working with the same interface kind.")
            raise SystemExit(1)

    @staticmethod
    def _process_driver_obj(obj):
        """Process a loaded interface driver object.

        :param obj: n_interface.LinuxInterfaceDriver obj.
        :return: An augmented object containing _interface_kind_ attribute that
                 represents the kind of interface a driver creates.
        """

        LOG.debug("Processing driver object of type:%s", obj.__class__)
        if (obj.__class__ is n_interface.OVSInterfaceDriver and
                obj.conf.ovs_use_veth):
            interface_kind = constants.INTERFACE_KIND_VETH
        else:
            interface_kind = (
                MultiInterfaceDriver.driver_to_interface_kind.get(
                    obj.__class__, constants.INTERFACE_KIND_UNKNOWN))

        if hasattr(obj, '_interface_kind_'):
            LOG.error("Attribute '_interface_kind_' defined, unexpected err.")
            raise SystemExit(1)
        obj._interface_kind_ = interface_kind
        return obj

    @staticmethod
    def load_interface_driver_mappings(conf, **kwargs):
        """Load interface drivers for agents like DHCP or L3 agent.

        :param conf: driver configuration object
        :param kwargs: additional keyword arguments
        :raises SystemExit of 1 if drivers cannot be loaded
        """

        try:
            driver_mapping = {}
            mappings = conf.multi_interface_driver_mappings.split(',')
            mappings = helpers.parse_mappings(mappings, False)
            for physnet, intf_driver_name in mappings.items():
                loaded_class = runtime.load_class_by_alias_or_classname(
                    n_agent_utils.INTERFACE_NAMESPACE, intf_driver_name)
                obj = MultiInterfaceDriver._process_driver_obj(
                    loaded_class(conf, **kwargs))
                driver_mapping[physnet] = obj
            MultiInterfaceDriver._check_drivers(driver_mapping)
            return driver_mapping
        except ImportError:
            LOG.error("Error loading interface driver '%s'",
                      conf.interface_driver)
            raise SystemExit(1)

    @classmethod
    def _init_network_cache(cls, conf, get_networks_cb, fields):
        """Initialize a global network cache to be shared between all Multi
        interface driver instances.

        :param conf: configuration object
        :param get_networks_cb: RPC callback to get network information
        :param fields: a list of fields to retrieve per network
        """
        if cls.network_cache:
            # Already initialized
            return
        with cls._cache_init_lock:
            if cls.network_cache:
                # Double checking if already initialized by someone else
                return
            cls.network_cache = network_cache.SafeNetworkCache(
                get_networks_cb, fields)
            # Populate cache
            cls.network_cache.refresh()
            if conf.enable_multi_interface_driver_cache_maintenance:
                remove_stale_entries_loop = (
                    loopingcall.FixedIntervalLoopingCall(
                        cls.network_cache.remove_stale_networks))
                remove_stale_entries_loop.start(
                    DEFAULT_CACHE_CLEANING_INTERVAL,
                    initial_delay=DEFAULT_CACHE_CLEANING_INTERVAL)

    @classmethod
    def _get_networks_from_cache(cls, fields=None, filters=None):
        nets = []
        if cls.network_cache is None:
            LOG.Error("MultiInterfaceDriver network cache was not initialized")
            return nets

        # NOTE(adrianc): since we want to work solely against cls.network_cache
        # it is required to impose some restrictions on filters and fields.
        if filters and list(six.iterkeys(filters)) != ['id']:
            LOG.error("Cache backed get_networks() does not support provided "
                      "filters: %s", filters)
            return nets

        if fields and not set(fields).issubset(set(
                cls.network_cache.network_fields)):
            LOG.warning("Cache backed get_networks() does not contain all "
                        "provided fields: %s", fields)

        if filters is None:
            nets = cls.network_cache.get_all()
        else:
            ids = filters.get('id')
            for id in ids:
                nets.append(cls.network_cache.get(id))

        if fields:
            nets = [{field: net[field] for field in fields
                if field in net.keys()} for net in nets]
            # In case of an empty entry due to missing fields
            # remove from response.
            nets = [net for net in nets if len(net)]

        return nets

    def _get_driver_for_existing_interface(self, device_name, namespace=None):
        dev = n_ip_lib.IPDevice(device_name, namespace)
        kind = dev.link.link_kind
        for driver in six.itervalues(self.drivers):
            if driver._interface_kind_ == kind:
                return driver
        LOG.error("No interface driver found for interface %s of kind %s",
                  device_name, kind)
        return None

    def plug_new(self, network_id, port_id, device_name, mac_address,
                 bridge=None, namespace=None, prefix=None, mtu=None):
        """Plugin the interface."""
        network = MultiInterfaceDriver.network_cache.get(network_id)
        physnet = network.get(constants.PHYSICAL_NETWORK)
        if physnet is None:
            # Tunnels(e.g vxlan, gre) map to the 'nil' interface driver.
            physnet = 'nil'
        try:
            driver = self.drivers[physnet]
            driver.plug_new(network_id, port_id, device_name, mac_address,
                            bridge, namespace, prefix, mtu)
        except KeyError:
            LOG.error("Interface driver not found for physnet: %s", physnet)

    def unplug(self, device_name, bridge=None, namespace=None, prefix=None):
        """Unplug the interface."""
        if not n_ip_lib.device_exists(device_name, namespace):
            # Nothing to unplug.
            return

        driver = self._get_driver_for_existing_interface(device_name,
                                                         namespace)
        if driver:
            driver.unplug(device_name, bridge, namespace, prefix)
        else:
            LOG.error("Failed to unplug interface %s, did not find matching "
                      "interface driver to unplug interface.", device_name)

    def set_mtu(self, device_name, mtu, namespace=None, prefix=None):
        """Set interface MTU"""
        driver = self._get_driver_for_existing_interface(device_name,
                                                         namespace)
        if driver:
            driver.set_mtu(device_name, mtu, namespace, prefix)
        else:
            LOG.error("Failed to set MTU for %s, did not find matching "
                      "interface driver for interface.", device_name)



#!/usr/bin/env python
# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import subprocess

import mock
import six

from networking_mlnx._i18n import _LE
from networking_mlnx.eswitchd.utils import pci_utils
from networking_mlnx.tests import base

if six.PY3:
    @contextlib.contextmanager
    def nested(*contexts):
        with contextlib.ExitStack() as stack:
            yield [stack.enter_context(c) for c in contexts]
else:
    nested = contextlib.nested


class TestPciUtils(base.TestCase):

    def setUp(self):
        super(TestPciUtils, self).setUp()
        self.pci_utils = pci_utils.pciUtils()

    def test_get_vfs_info_not_found_device(self):
        pf = "pf_that_does_not_exist"
        with mock.patch.object(pci_utils, 'LOG') as LOG:
            self.pci_utils.get_vfs_info(pf)
            LOG.error.assert_called_with(_LE("PCI device %s not found"), pf)

    def test_get_dev_attr_valid_attr(self):
        cmd = "find /sys/class/net/*/device/vendor | head -1 | cut -d '/' -f5"
        pf = subprocess.check_output(cmd, shell=True)
        pf = pf.strip().decode("utf-8")
        if pf:
            attr_path = "/sys/class/net/%s/device/vendor" % pf
            return_val = self.pci_utils.get_dev_attr(attr_path)
            self.assertIsNotNone(return_val)

    def test_get_dev_attr_invalid_attr(self):
        attr_path = "/path/that/does/not/exist"
        return_val = self.pci_utils.get_dev_attr(attr_path)
        self.assertIsNone(return_val)

    def test_verify_vendor_pf_valid_vendor(self):
        cmd = "find /sys/class/net/*/device/vendor | head -1 | cut -d '/' -f5"
        pf = subprocess.check_output(cmd, shell=True)
        pf = pf.strip().decode("utf-8")
        if pf:
            attr_path = "/sys/class/net/%s/device/vendor" % pf
            attr = subprocess.check_output("cat %s" % attr_path, shell=True)
            attr = attr.strip().decode("utf-8")
            return_val = self.pci_utils.verify_vendor_pf(pf, attr)
            self.assertTrue(return_val)

    def test_verify_vendor_pf_invalid_vendor(self):
        cmd = "ls -U /sys/class/net | head -1"
        pf = subprocess.check_output(cmd, shell=True)
        pf = pf.strip()
        attr = "0x0000"

        return_val = self.pci_utils.verify_vendor_pf(pf, attr)
        self.assertFalse(return_val)

    def test_is_sriov_pf_false(self):
        pf = "pf_that_does_not_exist"

        is_sriov = self.pci_utils.is_sriov_pf(pf)
        self.assertFalse(is_sriov)



# Copyright 2016 Mellanox Technologies, Ltd
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils

from networking_mlnx.db import db
from networking_mlnx.plugins.ml2.drivers.sdn import constants as sdn_const


def _is_valid_operation(session, row):
    # Check if there are older updates in the queue
    if db.check_for_older_ops(session, row):
        return False
    return True


def validate_network_operation(session, row):
    """Validate the network operation based on dependencies.

    Validate network operation depending on whether it's dependencies
    are still in 'pending' or 'processing' state. e.g.
    """
    if row.operation == sdn_const.DELETE:
        # Check for any pending or processing create or update
        # ops on this uuid itself
        if db.check_for_pending_or_processing_ops(
            session, row.object_uuid, [sdn_const.PUT,
                                       sdn_const.POST]):
            return False
        if db.check_for_pending_delete_ops_with_parent(
            session, sdn_const.PORT, row.object_uuid):
            return False
    elif (row.operation == sdn_const.PUT and
            not _is_valid_operation(session, row)):
        return False
    return True


def validate_port_operation(session, row):
    """Validate port operation based on dependencies.

    Validate port operation depending on whether it's dependencies
    are still in 'pending' or 'processing' state. e.g.
    """
    if row.operation in (sdn_const.POST, sdn_const.PUT):
        network_dict = jsonutils.loads(row.data)
        network_id = network_dict['network_id']
        # Check for pending or processing network operations
        ops = db.check_for_pending_or_processing_ops(
            session, network_id, [sdn_const.POST])
        if ops:
            return False
    return _is_valid_operation(session, row)


_VALIDATION_MAP = {
    sdn_const.NETWORK: validate_network_operation,
    sdn_const.PORT: validate_port_operation,
}


def validate(session, row):
    """Validate resource dependency in journaled operations.

    :param session: db session
    :param row: entry in journal entry to be validated
    """
    return _VALIDATION_MAP[row.object_type](session, row)


def register_validator(object_type, validator):
    """Register validator function for given resource.

    :param object_type: neutron resource type
    :param validator: function to be registered which validates resource
         dependencies
    """
    _VALIDATION_MAP[object_type] = validator



# Copyright 2016 Mellanox Technologies, Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""adding sdn journal db

Revision ID: 9f30890cfbd1
Create Date: 2016-08-07 10:57:15.895551

"""

from alembic import op
import sqlalchemy as sa

from networking_mlnx.plugins.ml2.drivers.sdn import constants as sdn_const

# revision identifiers, used by Alembic.
revision = '9f30890cfbd1'
down_revision = '65b6db113427b9'


def upgrade():
    op.create_table(
        'sdn_journal',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('object_type', sa.String(length=36), nullable=False),
        sa.Column('object_uuid', sa.String(length=36), nullable=False),
        sa.Column('operation', sa.String(length=36), nullable=False),
        sa.Column('data', sa.PickleType(), nullable=True),
        sa.Column('job_id', sa.String(length=36), nullable=True),
        sa.Column('state',
                  sa.Enum(sdn_const.PENDING, sdn_const.FAILED,
                          sdn_const.PROCESSING, sdn_const.COMPLETED,
                          name='state'),
                  nullable=False, default=sdn_const.PENDING),
        sa.Column('retry_count', sa.Integer, default=0),
        sa.Column('created_at', sa.DateTime, default=sa.func.now()),
        sa.Column('last_retried', sa.TIMESTAMP, server_default=sa.func.now(),
                  onupdate=sa.func.now())
    )
